
module Plugins.Equality where

import Text.PrettyPrint
import Control.Monad       ( when )
import Control.Monad.Trans ( liftIO )

import Syntax.Internal
import Syntax.Pretty

import TypeChecker
import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Options
import TypeChecker.Computation
import TypeChecker.UnfoldConstants
import TypeChecker.Abstraction
import TypeChecker.Values
import TypeChecker.ValueToTerm
import TypeChecker.Terms

import Utils.Error
import Utils.Monads
import Utils.Names

import Interface.Options

import Plugins

eq_Eq		= "Eq"
eq_congruence	= "congruence"
eq_symmetry	= "symmetry"

plugin :: SomePlugin
plugin = mkPlugin
    Plugin  { plName		= "eq"
            , plHelp            = show (fsep $ pwords "Does automatic symmetry and substitutivity in equality proofs.")
				    ++ "\n"
	    , plInit		= \_ -> return ()
	    , plFinal		= \_ _ -> return ()
	    , plCheck		= checkEquality
	    , plInfer		= \_ _ _ -> failure noInferenceError
	    , plOptions		= []
	    , plDefaultOptions	= ()
	    }

checkEquality :: () -> () -> [Term] -> Value -> TC (Value, ())
checkEquality _ _ [proof] t =
    sayWhen (whenSearchingForProof proof t) $
    do	(b,v,w)	    <- isEq t
	(a,v',w')   <- isEq =<< termType =<< inferType proof
	let solutions =
		[ (False, f) | f <- abstract v' v, f w' === w ] ++
		[ (True,  g) | g <- abstract w' v, g v' === w ]
	case solutions of
	    []	    -> failure noSolutionFoundError
	    sol:_   ->
		-- If we find a solution we build the proof term
		-- and run it through the typechecker (just to be sure).
		do  e <- buildProof a b v' w' sol
		    t' <- checkType e t

		    -- Debug printing
		    verbose <- optVerbose <$> commandLineOptions
		    when (verbose >= 2) $
			liftIO $ print $
			    vcat[ fsep  [ text "+++ eq-plugin proved"
					, nest 4 $ pretty (unfoldAllConstants t)
					]
				, nest 4 $ text ":" <+> pretty e
				]

		    v <- evaluateTypedTerm t'
		    return (v, ())
    where
	buildProof a b x y (useSym, f) =
	    do	a'  <- valueToTerm a
		b'  <- valueToTerm b
		x'  <- valueToTerm x
		y'  <- valueToTerm y
		f'  <- mkLam a f
		let xy = proof
		    yx = sym `apps` [ a', x', y', xy ]
		return $
		    if useSym
			then mkProof a' b' f' y' x' yx
			else mkProof a' b' f' x' y' xy
	    where
		cong	= Def eq_congruence ()
		sym	= Def eq_symmetry ()
		apps	= foldl App
		mkProof t u e x y xy
		    | isId a f	= xy
		    | otherwise	= cong `apps` [ t, u, e, x, y, xy ]

	isId a f = x === f x
	    where x = VId "<any>" a

	mkLam v f =
	    do	x <- fresh "x"
		b <- valueToTerm (f $ VId x v)
		return $ Lam x b ()

	isEq v =
	    case vAppView $ unfoldConstants v of
		VAVDef c _ [a,v,w]
		    | c == eq_Eq    -> return (a,v,w)
		_		    -> failure $ onlyEqualityProofsError v
checkEquality _ _ _ _ =
    failure wrongNumberOfArgumentsError

whenSearchingForProof p v =
    fsep $ pwords "searching for proof of" ++ [pretty v] ++
	   pwords "using" ++ [pretty p]

noInferenceError :: ErrorFun
noInferenceError e =
    e	{ errWhat   = fsep $ pwords "type inference failed"
	, errWhy    = fsep $ pwords "the type of an automatic equality proof cannot be inferred"
	}

onlyEqualityProofsError :: Value -> ErrorFun
onlyEqualityProofsError v e =
    e	{ errWhat   = fsep $ pwords "don't know what to do with" ++ [pretty v]
	, errWhy    = fsep $ pwords "it is not an equality"
	}

noSolutionFoundError :: ErrorFun
noSolutionFoundError e =
    e	{ errWhat   = fsep $ pwords "proof failed"
	, errWhy    = fsep $ pwords $
			"it wasn't a simple application of symmetry " ++
			"and/or congruence (or there is a bug in " ++
			"the proof search)"
	}

wrongNumberOfArgumentsError :: ErrorFun
wrongNumberOfArgumentsError e =
    e	{ errWhat   = fsep $ pwords "expected a single argument to eq-plugin"
	}

