{-# OPTIONS -fglasgow-exts #-}

module Syntax.Values
    ( Value -- abstract
    , vId, vBId, vDef, vCon, vApp, vPi, vLam, vSort, vTop, vPair, vProj, vSigma, vMeta, vStruct
    , vDefToCon
    , MetaRef, MetaVar(..)
    , SortValue(..)
    , ProjectionValue(..)
    , DefValue(..), ActualDefinition(..), ClauseValue(..)
    , TelescopeValue, Type, el, unEl, getSort, typeMap, typeMapM
    , Abs -- abstract
    , absName, absBody, mkAbs, absMap, absMapM
    , ValueView(..), vView, unView
    , ValueAppView(..), vAppView, unAppView
    , ValueAppProjView(..), AppOrProj(..), vAppProjView
    , showVal, FreeVars
    , getClausesTC, getClausesRT
    , mapClauses
    )
    where

import Control.Monad.Trans
import Data.Generics
import Data.IORef
import Data.List

import Syntax.Internal
import Syntax.Name

import Utils.Monads

-- * Values 
----------------------------------------------------------------------

-- | The internal representation of values. Accessed by the views.
data Value  = VId' Name
              -- ^ . @x :: V@
	    | VBId' Int
	      -- ^ . deBruijn variable
	    | VDef' Name Int	    -- ^ defined name
	    | VCon' Name Int Arity  -- ^ constructor
	    | VApp' Value Value   
              -- ^ . @v w@ where @v@ is not a @VLam@
	    | VPi' Type (Abs Type)
              -- ^ . @(Fun W F) :: V@, type value @V@ must be a sort
	    | VLam' (Abs Value)
              -- ^ . @(Lam F) :: V@, type value @V@ must be a @Fun@
	    | VSort' SortValue
	      -- ^ . user sort or @#n@.
	    | VTop'
              -- ^ . @Top :: V@
	    | VPair' Value Value
	      -- ^ . @(v,w) :: V@, @V@ must be a @Sigma@
	    | VProj' Value ProjectionValue
	      -- ^ . @v.p :: V@ Invariant: No named projections.
	    | VSigma' Type (Abs Type)
	      -- ^ . @(Sigma W F) :: V@, @V@ must be a sort
	    | VMeta' MetaRef
	      -- ^ Meta variable
	    | VStruct' [(String,(Hidden,Value))]
	      -- ^ . @struct x1 = e1; ..; xn = en@ Structs are flat.
    deriving (Typeable) --, Data)

-- * Constructors ---------------------------------------------------------
---------------------------------------------------------------------------

vId	= VId'
vBId	= VBId'
vDef	= VDef'
vCon	= VCon'
vApp	= VApp'
vPi	= VPi'
vLam	= VLam'
vSort	= VSort'
vTop	= VTop'
vPair	= VPair'
vProj	= VProj'
vSigma	= VSigma'
vMeta	= VMeta'
vStruct = VStruct'

-- | Make any top-level 'vDef' into a 'vCon'.
vDefToCon :: Value -> Value
vDefToCon v =
    case vAppView v of
	VAVDef x h vs	-> foldl vApp (vCon x h h) vs
	_		-> v

-- | The official view of a value.
data ValueView
	    = VId Name
	    | VBId Int
	    | VDef Name Int
	    | VCon Name Int Arity
	    | VApp Value Value
	    | VPi Type (Abs Type)
	    | VLam (Abs Value)
	    | VSort SortValue
	    | VTop
	    | VPair Value Value
	    | VProj Value ProjectionValue
	    | VSigma Type (Abs Type)
	    | VMeta MetaRef
	    | VStruct [(String, (Hidden,Value))]
    deriving (Typeable) --, Data)

type MetaRef = IORef MetaVar

data MetaVar = Open
	     | Instantiated Value
    deriving (Typeable) --, Data)

data SortValue = VSISet
	       | VSProp
	       | VSUniv Universe
    deriving (Typeable) --, Data)

data ProjectionValue = VFst | VSnd | VField Hidden String
    deriving (Eq, Typeable) --, Data)

-- | This should be fine
type TelescopeValue = [Type]

data Type = El	{ unEl	  :: Value
		, getSort :: SortValue
		}
    deriving (Typeable)

el :: SortValue -> Value -> Type
el s v = El v s

-- | The function should preserve the sort.
typeMap :: (Value -> Value) -> Type -> Type
typeMap f (El v s) = El (f v) s

-- | The function should preserve the sort.
typeMapM :: Monad m => (Value -> m Value) -> Type -> m Type
typeMapM f (El v s) = flip El s <$> f v

instance Eq Value where
    VId' x		== VId' y		= x == y
    VBId' n		== VBId' m		= n == m
    VDef' x _		== VDef' y _		= x == y
    VCon' x _ _		== VCon' y _ _		= x == y
    VApp' v w		== VApp' v' w'		= (v,w) == (v',w')
    VPi' v f		== VPi' w f'		= (v,f) == (w,f')
    VLam' f		== VLam' f'		= f == f' 
    VSort' VSISet	== VSort' VSISet	= True
    VSort' VSProp	== VSort' VSProp	= True
    VSort' (VSUniv u)	== VSort' (VSUniv u')	= u == u'
    VTop'		== VTop'		= True
    VPair' v w		== VPair' v' w'		= (v,w) == (v',w')
    VProj' v p		== VProj' w q		= (v,p) == (w,q)
    VSigma' v f		== VSigma' w f'		= (v,f) == (w,f')
    VMeta' r		== VMeta' r'		= r == r'
    VStruct' xs		== VStruct' ys		= xs == ys
    _			== _			= False

instance Eq Type where
    El v _ == El w _ = v == w

instance Eq a => Eq (Abs a) where
    Abs _ b == Abs _ b' = b == b'

{------------------------------------------------------------------------
    Abstractions
 ------------------------------------------------------------------------}

data Abs a = Abs { absName :: String	-- just a suggestion
		 , absBody :: a
		 }
    deriving (Typeable) --(Data, Typeable)

-- | Takes the type of the argument.
mkAbs :: String -> a -> Abs a
mkAbs = Abs

instance Functor Abs where
    fmap f (Abs x b) = Abs x (f b)

absMap :: (a -> b) -> Abs a -> Abs b
absMap = fmap

absMapM :: Monad m => (a -> m b) -> Abs a -> m (Abs b)
absMapM f (Abs x b) = Abs x <$> f b

{------------------------------------------------------------------------
    Definition values
 ------------------------------------------------------------------------}

type FreeVars = Int

-- | Maybe we want different kinds of DefValues?
data DefValue = DefVal Hidden Arity FreeVars ActualDefinition Type
    deriving (Typeable) --, Data)

data ActualDefinition = ConcrDef [ClauseValue]
		      | AbstrDef [ClauseValue]
    deriving (Typeable)

-- | Invariant: If @ClauseVal ps v@ and @xs@ are the variables of
--   @ps@, then @v@ should have the form @\xs -> v'@.
data ClauseValue    = ClauseVal [Pattern] Value
    deriving (Typeable) --, Data)

mapClauses :: ([ClauseValue] -> [ClauseValue]) -> ActualDefinition -> ActualDefinition
mapClauses f (AbstrDef cl) = AbstrDef $ f cl
mapClauses f (ConcrDef cl) = ConcrDef $ f cl

mapClausesM :: Monad m => ([ClauseValue] -> m [ClauseValue]) -> ActualDefinition -> m ActualDefinition
mapClausesM f (AbstrDef cl) = AbstrDef <$> f cl
mapClausesM f (ConcrDef cl) = ConcrDef <$> f cl

getClausesTC :: ActualDefinition -> [ClauseValue]
getClausesTC (AbstrDef _)  = []
getClausesTC (ConcrDef cl) = cl

getClausesRT :: ActualDefinition -> [ClauseValue]
getClausesRT (AbstrDef cl) = cl
getClausesRT (ConcrDef cl) = cl

-- * Views on values 
----------------------------------------------------------------------

data AppOrProj	= IsApp Value
		| IsProj ProjectionValue

data ValueAppProjView
	= VAPVId Name [AppOrProj]
	| VAPVBId Int [AppOrProj]
	| VAPVMeta MetaRef [AppOrProj]
	| VAPVDef Name Int [AppOrProj]
	| VAPVCon Name Int Arity [AppOrProj]
	| VAPVTop
	| VAPVSort SortValue
	| VAPVLam (Abs Value)
	| VAPVPi Type (Abs Type)
	| VAPVPair Value Value
	| VAPVSigma Type (Abs Type)
	| VAPVStruct [(String, (Hidden,Value))]

data ValueAppView
	= VAVId Name [Value]
	| VAVBId Int [Value]
	| VAVMeta MetaRef [Value]
	| VAVDef Name Int [Value]
	| VAVCon Name Int Arity [Value]
	| VAVProj Value ProjectionValue [Value]
	| VAVTop
	| VAVSort SortValue
	| VAVLam (Abs Value)
	| VAVPi Type (Abs Type)
	| VAVPair Value Value
	| VAVSigma Type (Abs Type)
	| VAVStruct [(String, (Hidden,Value))]

vView :: Value -> ValueView
vView v =
    case v of
	VId' x		-> VId x
	VBId' n		-> VBId n
	VMeta' m	-> VMeta m
	VDef' x h	-> VDef x h
	VCon' x h ar	-> VCon x h ar
	VPi' v b	-> VPi v b
	VLam' b		-> VLam b
	VSort' s	-> VSort s
	VTop'		-> VTop
	VPair' v w	-> VPair v w
	VSigma' w b	-> VSigma w b
	VApp' v w	-> VApp v w
	VProj' v p	-> VProj v p
	VStruct' xs	-> VStruct xs

unView :: ValueView -> Value
unView v =
    case v of
	VId x		-> VId' x
	VBId n		-> VBId' n
	VMeta m		-> VMeta' m
	VDef x h	-> VDef' x h
	VCon x h ar	-> VCon' x h ar
	VPi v cl	-> VPi' v cl
	VLam cl		-> VLam' cl
	VSort s		-> VSort' s
	VTop		-> VTop'
	VPair v w	-> VPair' v w
	VSigma w f	-> VSigma' w f
	VApp v w	-> VApp' v w
	VProj v p	-> VProj' v p
	VStruct xs	-> VStruct' xs

vAppProjView :: Value -> ValueAppProjView
vAppProjView v =
    case v of
	VId' x		-> VAPVId x []
	VBId' n		-> VAPVBId n []
	VMeta' m	-> VAPVMeta m []
	VDef' x h	-> VAPVDef x h []
	VCon' x h ar	-> VAPVCon x h ar []
	VPi' v cl	-> VAPVPi v cl
	VLam' cl	-> VAPVLam cl
	VSort' s	-> VAPVSort s
	VTop'		-> VAPVTop
	VPair' v w	-> VAPVPair v w
	VSigma' w f	-> VAPVSigma w f
	VApp' v w	-> vapvApp v (vAppProjView v) w
	VProj' v p	-> vapvProj v (vAppProjView v) p
	VStruct' xs	-> VAPVStruct xs

vAppView :: Value -> ValueAppView
vAppView v =
    case v of
	VId' x		-> VAVId x []
	VBId' n		-> VAVBId n []
	VMeta' m	-> VAVMeta m []
	VDef' x h	-> VAVDef x h []
	VCon' x h ar	-> VAVCon x h ar []
	VProj' v p	-> VAVProj v p []
	VPi' v cl	-> VAVPi v cl
	VLam' cl	-> VAVLam cl
	VSort' u	-> VAVSort u
	VTop'		-> VAVTop
	VPair' v w	-> VAVPair v w
	VSigma' w f	-> VAVSigma w f
	VApp' v w	-> vavApp (vAppView v) w
	VStruct' xs	-> VAVStruct xs

unAppView :: ValueAppView -> Value
unAppView v =
    case v of
	VAVId x vs	    -> vId x `vApps` vs
	VAVBId n vs	    -> vBId n `vApps` vs
	VAVMeta m vs	    -> vMeta m `vApps` vs
	VAVDef x h vs	    -> vDef x h `vApps` vs
	VAVCon x h ar vs    -> vCon x h ar `vApps` vs
	VAVProj v p vs	    -> vProj v p `vApps` vs
	VAVPi v cl	    -> vPi v cl
	VAVLam cl	    -> vLam cl
	VAVSort u	    -> vSort u
	VAVTop		    -> vTop
	VAVPair v w	    -> vPair v w
	VAVSigma w f	    -> vSigma w f
	VAVStruct xs	    -> vStruct xs
    where
	vApps = foldl vApp

vavApp :: ValueAppView -> Value -> ValueAppView
vavApp v w =
    case v of
	VAVId x vs	    -> VAVId x (vs ++ [w])
	VAVBId x vs	    -> VAVBId x (vs ++ [w])
	VAVMeta m vs	    -> VAVMeta m (vs ++ [w])
	VAVDef x h vs	    -> VAVDef x h (vs ++ [w])
	VAVCon x h ar vs    -> VAVCon x h ar (vs ++ [w])
	VAVProj v p vs	    -> VAVProj v p (vs ++ [w])
	_		    -> error "Internal error in vavApp!"

vapvApp :: Value -> ValueAppProjView -> Value -> ValueAppProjView
vapvApp v' v w =
    case v of
	VAPVId x vs	    -> VAPVId x (vs ++ [IsApp w])
	VAPVBId x vs	    -> VAPVBId x (vs ++ [IsApp w])
	VAPVMeta m vs	    -> VAPVMeta m (vs ++ [IsApp w])
	VAPVDef x h vs	    -> VAPVDef x h (vs ++ [IsApp w])
	VAPVCon x h ar vs   -> VAPVCon x h ar (vs ++ [IsApp w])
	_		    -> error $ "Internal error in vapvApp: " ++ showVal (vApp v' w)

vapvProj :: Value -> ValueAppProjView -> ProjectionValue -> ValueAppProjView
vapvProj v' v p =
    case v of
	VAPVId x vs	    -> VAPVId x (vs ++ [IsProj p])
	VAPVBId x vs	    -> VAPVBId x (vs ++ [IsProj p])
	VAPVMeta m vs	    -> VAPVMeta m (vs ++ [IsProj p])
	VAPVDef x h vs	    -> VAPVDef x h (vs ++ [IsProj p])
	VAPVCon x h ar vs   -> VAPVCon x h ar (vs ++ [IsProj p])
	_		    -> error $ "Internal error in vapvProj: " ++ showVal (vProj v' p)

-- For debugging purposes
showVal :: Value -> String
showVal v =
    case v of
	VId' x	    -> nameString x
	VBId' n	    -> "@" ++ show n
	VDef' x _   -> nameString x
	VCon' x _ _ -> nameString x
	VPi' v cl   ->
	    concat  ["((", absName cl, " : ", showVal (unEl v), ")"
		    , " -> "
		    , showVal (unEl $ absBody cl)
		    , ")"
		    ]
	VLam' cl    ->
	    concat  ["(\\", absName cl, " -> "
		    , showVal (absBody cl)
		    , ")"
		    ]
	VSort' VSISet	    -> "ISet"
	VSort' (VSUniv u)   -> "#" ++ show (uValue u)
	VSort' VSProp	    -> "Prop"
	VTop'		    -> "Top"
	VPair' v w	    ->
	    concat  [ "(", showVal v, ", ", showVal w, ")" ]
	VSigma' w f	    ->
	    concat  [ "{", absName f, " : ", showVal (unEl w), "; "
		    , showVal (unEl $ absBody f)
		    , "}"
		    ]
	VProj' v p	    ->
	    concat  [ showVal v, ".", case p of
					VFst -> "1"
					VSnd -> "2"
					VField h x -> x
		    ]
	VApp' v w	    ->
	    concat  [ "(", showVal v, " ", showVal w, ")" ]
	VMeta' m 	    -> "?"
	VStruct' xs -> "(struct " ++ concat (intersperse "; " $ map f xs) ++ ")"
	    where
		f (x,(h,v)) = x ++ "[" ++ show h ++ "] = " ++ showVal v

